;;; hmac.el --- Hashed Message Authentication Codes

;;; Copyright (C) 1998 Simon Josefsson <jas@pdc.kth.se>

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

;;; Commentary:

;;; This is a quick'n'dirty implementation of RFC2104.
;;;
;;; Example:
;;;
;;; (require 'md5)
;;; (hmac 'md5 64 16 "Jefe" "what do ya want for nothing?")
;;; "750c783e6ab0b503eaa86e310a5db738"
;;;
;;; 64 is block length of hash function (64 for MD5 and SHA), 16 is
;;; resulting hash length (16 for MD5, 20 for SHA).
;;;
;;; Tested with FSF Emacs 20.2 and XEmacs 20.3. Does not work with
;;; FSF Emacs 19.
;;;

(require 'cl)

;; 0x36 == 54 == '6'
(defconst hmac-ipad ?\x36)

;; 0x5C == 92 == '\'
(defconst hmac-opad ?\x5C)

;; 0x00
(defconst hmac-zero ?\x00)

(if (not (fboundp 'hexl-hex-string-to-integer))
    (defun hexl-hex-string-to-integer (string)
      (string-to-number string 16)))

(if (not (fboundp 'string-to-char-list))
    (defun string-to-char-list (string)
      "Return a list of which elements are characters in the STRING."
      (let* ((len (length string))
	     (i 0)
	     l chr)
	(while (< i len)
	  (setq chr (string-to-char (substring string i)))
	  (setq l (cons chr l))
	  (setq i (1+ i)))
	(nreverse l))))

(defun hmac (hash block-length hash-length key text)
  (let* (;; if key is longer than B, reset it to HASH(key)
	 (key (if (> (length key) block-length) 
		  (funcall hash key) key))
	 (k_ipad (string-to-char-list key))
	 (k_opad (string-to-char-list key)))
    ;; zero pad k_ipad/k_opad
    (while (< (length k_ipad) block-length)
      (setq k_ipad (append k_ipad (list hmac-zero))))
    (while (< (length k_opad) block-length)
      (setq k_opad (append k_opad (list hmac-zero))))
    ;; XOR key with ipad/opad into k_ipad/k_opad
    (setq k_ipad (mapcar (lambda (c) (logxor c hmac-ipad)) k_ipad))
    (setq k_opad (mapcar (lambda (c) (logxor c hmac-opad)) k_opad))
    ;; perform inner hash
    (let ((first-round (funcall hash (concat k_ipad text)))
	  de-hexed)
      (while (< 0 (length first-round))
	(push (hexl-hex-string-to-integer (substring first-round -2)) de-hexed)
	(setq first-round (substring first-round 0 -2)))
      ;; perform outer hash
      (let ((second-round 
	     (funcall hash (concat k_opad de-hexed))))
	second-round))))


;  (let ((k_ipad (make-vector block-length hmac-zero))
;	(k_opad (make-vector block-length hmac-zero))
;	;; if key is longer than B, reset it to HASH(key)
;	(key (if (> (length key) block-length) 
;		 (funcall hash key) key)))
;    ;; copy in key into k_ipad/k_opad
;    (let ((j 0))
;      (while (< j (length key))
;	(aset k_ipad j (string-to-char (substring key j)))
;	(setq j (1+ j))))
;    (let ((j 0))
;      (while (< j (length key))
;	(aset k_opad j (string-to-char (substring key j)))
;	(setq j (1+ j))))
;    ;; XOR key with ipad/opad into k_ipad/k_opad
;    (setq k_ipad (mapvector (lambda (c) (logxor c hmac-ipad)) k_ipad))
;    (setq k_opad (mapvector (lambda (c) (logxor c hmac-opad)) k_opad))
;    ;; perform inner hash
;    (let ((first-round
;	   (funcall hash (concat k_ipad text))))
;      ;; perform outer hash
;      (let ((second-round 
;	     (funcall hash (concat k_opad first-round))))
;	second-round))))

(provide 'hmac)
